# -*- coding: utf-8 -*-

from typing import List, Optional, Set, Tuple

from django.apps import apps
from django.db import DEFAULT_DB_ALIAS, connections
from django.db.migrations.graph import Node
from django.db.migrations.loader import MigrationLoader


def all_migrations(
    database: str = DEFAULT_DB_ALIAS,
    app_names: Optional[List[str]] = None,
) -> List[Node]:
    """
    Returns the sorted list of migrations nodes.

    The order is the same as when migrations are applied.

    When you might need this function?
    When you are testing the migration order.

    For example, imagine that you have a direct dependency:
    ``main_app.0002_migration`` and ``other_app.0001_initial``
    where ``other_app.0001_initial`` relies on the model or field
    introduced in ``main_app.0002_migration``.

    You can use ``dependencies`` field
    to ensure that everything works correctly.

    But, sometimes migrations are squashed,
    sometimes they are renamed, refactored, and moved.

    It would be better to have a test that will ensure
    that ``other_app.0001_initial`` comes after ``main_app.0002_migration``.
    And everything works as expected.
    """
    loader = MigrationLoader(connections[database])

    if app_names:
        _validate_app_names(app_names)
        targets = [
            key
            for key in loader.graph.leaf_nodes()
            if key[0] in app_names
        ]
    else:
        targets = loader.graph.leaf_nodes()
    return _generate_plan(targets, loader)


def nodes_to_tuples(nodes: List[Node]) -> List[Tuple[str, str]]:
    """Utility function to transform nodes to tuples."""
    return [
        (node[0], node[1])
        for node in nodes
    ]


def _validate_app_names(app_names: List[str]) -> None:
    """
    Validates the provided app names.

    Raises ```LookupError`` when incorrect app names are provided.
    """
    for app_name in app_names:
        apps.get_app_config(app_name)


def _generate_plan(
    targets: List[Node],
    loader: MigrationLoader,
) -> List[Node]:
    plan = []
    seen: Set[Node] = set()

    # Generate the plan
    for target in targets:
        for migration in loader.graph.forwards_plan(target):
            if migration not in seen:
                node = loader.graph.node_map[migration]
                plan.append(node)
                seen.add(migration)
    return plan



# -*- coding: utf-8 -*-

import pytest
from django.db import DEFAULT_DB_ALIAS


@pytest.fixture()
def migrator_factory(request, transactional_db):
    """
    Pytest fixture to create migrators inside the pytest tests.

    How? Here's an example.

    .. code:: python

        @pytest.mark.django_db
        def test_migration(migrator_factory):
            migrator = migrator_factory('custom_db_alias')
            old_state = migrator.before(('main_app', None))
            new_state = migrator.after(('main_app', '0001_initial'))

            assert isinstance(old_state, ProjectState)
            assert isinstance(new_state, ProjectState)

    Why do we import :class:`Migrator` inside the fixture function?
    Otherwise, coverage won't work correctly during our internal tests.
    Why? Because modules in Python are singletons.
    Once imported, they will be stored in memory and reused.

    That's why we cannot import ``Migrator`` on a module level.
    Because it won't be caught be coverage later on.
    """
    from django_test_migrations.migrator import Migrator  # noqa: WPS433

    def factory(database_name=None):
        migrator = Migrator(database_name)
        request.addfinalizer(migrator.reset)
        return migrator
    return factory


@pytest.fixture()
def migrator(migrator_factory):  # noqa: WPS442
    """
    Useful alias for ``'default'`` database in ``django``.

    That's a predefined instance of a ``migrator_factory``.

    How to use it? Here's an example.

    .. code:: python

        @pytest.mark.django_db
        def test_migration(migrator):
            old_state = migrator.before(('main_app', None))
            new_state = migrator.after(('main_app', '0001_initial'))

            assert isinstance(old_state, ProjectState)
            assert isinstance(new_state, ProjectState)

    Just one step easier than ``migrator_factory`` fixture.
    """
    return migrator_factory(DEFAULT_DB_ALIAS)



# -*- coding: utf-8 -*-

from typing import ClassVar, Optional

from django.db.migrations.state import ProjectState
from django.test import TransactionTestCase

from django_test_migrations.migrator import (  # noqa: WPS436
    Migrator,
    _Migration,
)


class MigratorTestCase(TransactionTestCase):
    """Used when using raw ``unitest`` library for test."""

    database_name: ClassVar[Optional[str]] = None
    old_state: ProjectState
    new_state: ProjectState

    #: Part of the end-user API. Used to tell what migrations we are using.
    migrate_from: ClassVar[_Migration]
    migrate_to: ClassVar[_Migration]

    def __init__(self, *args, **kwargs) -> None:
        """Initializes our :class:`Migrator` instance."""
        super().__init__(*args, **kwargs)

    def setUp(self) -> None:
        """
        Regular ``unittest`` styled setup case.

        What it does?
          - It starts with defining the initial migration state
          - Then it allows to run custom method
            to prepare some data before the migration will happen
          - Then it applies the migration and saves all states

        """
        super().setUp()
        self._pre_setup()
        self._migrator = Migrator(self.database_name)
        self.old_state = self._migrator.before(self.migrate_from)
        self.prepare()
        self.new_state = self._migrator.after(self.migrate_to)

    def prepare(self) -> None:
        """
        Part of the end-user API.

        Used to prepare some data before the migration process.
        """

    def tearDown(self) -> None:
        """Used to clean mess up after each test."""
        self._migrator.reset()
        super().tearDown()



# -*- coding: utf-8 -*-

from typing import Optional, Tuple, Union

from django.core.management import call_command
from django.db import DEFAULT_DB_ALIAS, connections
from django.db.migrations.executor import MigrationExecutor
from django.db.migrations.state import ProjectState

_Migration = Union[
    # Initial migration: None -> 0001
    Tuple[Optional[str], str],
    # Regular or rollback migration: 0001 -> 0002, or 0002 -> 0001
    Tuple[str, str],
    # Rollback migration to initial state: 0001 -> None
    Tuple[str, Optional[str]],
]


class Migrator(object):
    """
    Class to manage your migrations and app state.

    It is designed to be used inside the tests to ensure that migrations
    are working as intended: both data and schema migrations.

    This class can be but probably should not be used directly.
    Because we have utility test framework
    integrations for ``unitest`` and ``pytest``.

    Use them for better experience.
    """

    def __init__(
        self,
        database: Optional[str] = None,
    ) -> None:
        """That's where we initialize all required internals."""
        if database is None:
            database = DEFAULT_DB_ALIAS

        self._database: str = database
        self._executor = MigrationExecutor(connections[self._database])

    def before(self, migrate_from: _Migration) -> ProjectState:
        """Reverse back to the original migration."""
        return self._executor.migrate([migrate_from])

    def after(self, migrate_to: _Migration) -> ProjectState:
        """Apply the next migration."""
        self._executor.loader.build_graph()  # reload.
        return self._executor.migrate([migrate_to])

    def reset(self) -> None:
        """Reset the state to the most recent one."""
        return call_command('migrate')



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['django_test_migrations', 'django_test_migrations.contrib']

package_data = \
{'': ['*']}

entry_points = \
{'pytest11': ['django_test_migrations = '
              'django_test_migrations.contrib.pytest_plugin']}

setup_kwargs = {
    'name': 'django-test-migrations',
    'version': '0.1.0',
    'description': 'Test django schema and data migrations, including ordering',
    'long_description': '# django-test-migrations\n\n[![wemake.services](https://img.shields.io/badge/%20-wemake.services-green.svg?label=%20&logo=data%3Aimage%2Fpng%3Bbase64%2CiVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAMAAAAoLQ9TAAAABGdBTUEAALGPC%2FxhBQAAAAFzUkdCAK7OHOkAAAAbUExURQAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAP%2F%2F%2F5TvxDIAAAAIdFJOUwAjRA8xXANAL%2Bv0SAAAADNJREFUGNNjYCAIOJjRBdBFWMkVQeGzcHAwksJnAPPZGOGAASzPzAEHEGVsLExQwE7YswCb7AFZSF3bbAAAAABJRU5ErkJggg%3D%3D)](https://wemake.services)\n[![Build Status](https://travis-ci.com/wemake-services/django-test-migrations.svg?branch=master)](https://travis-ci.com/wemake-services/django-test-migrations)\n[![Coverage](https://coveralls.io/repos/github/wemake-services/django-test-migrations/badge.svg?branch=master)](https://coveralls.io/github/wemake-services/django-test-migrations?branch=master)\n[![Python Version](https://img.shields.io/pypi/pyversions/django-test-migrations.svg)](https://pypi.org/project/django-test-migrations/)\n[![wemake-python-styleguide](https://img.shields.io/badge/style-wemake-000000.svg)](https://github.com/wemake-services/wemake-python-styleguide)\n\n\n## Features\n\n- Allows to test `django` schema and data migrations\n- Allows to test both forward and rollback migrations\n- Allows to test the migrations order\n- Fully typed with annotations and checked with `mypy`, [PEP561 compatible](https://www.python.org/dev/peps/pep-0561/)\n- Easy to start: has lots of docs, tests, and tutorials\n\n\n## Installation\n\n```bash\npip install django-test-migrations\n```\n\nWe support several `django` versions:\n\n- `1.11`\n- `2.1`\n- `2.2`\n\nOther versions might work too, but they are not officially supported.\n\n\n## Testing django migrations\n\nTesting migrations is not a frequent thing in `django` land.\nBut, sometimes it is totally required. When?\n\nWhen we do complex schema or data changes\nand what to be sure that existing data won\'t be corrupted.\nWe might also want to be sure that all migrations can be safely rolled back.\nAnd as a final touch we want to be sure that migrations\nare in the correct order and have correct dependencies.\n\n### Testing forward migrations\n\nTo test all migrations we have a [`Migrator`](https://github.com/wemake-services/django-test-migrations/blob/master/django_test_migrations/migrator.py) class.\n\nIt has three methods to work with:\n\n- `.before()` which takes app and migration names to generate a state\n  before the actual migration happens\n- `.after()` which takes app and migration names to perform the actual migration\n- `.reset()` to clean everything up after we are done with testing\n\nSo, here\'s an example:\n\n```python\nfrom django_test_migrations.migrator import Migrator\n\nmigrator = Migrator(database=\'default\')\n\n# Initial migration, currently our model has only a single string field:\nold_state = migrator.before((\'main_app\', \'0001_initial\'))\nSomeItem = old_state.apps.get_model(\'main_app\', \'SomeItem\')\n\n# Let\'s create a model with just a single field specified:\nSomeItem.objects.create(string_field=\'a\')\nassert len(SomeItem._meta.get_fields()) == 2  # id + string_field\n\n# Now this migration will add `is_clean` field to the model:\nnew_state = migrator.after((\'main_app\', \'0002_someitem_is_clean\'))\nSomeItem = new_state.apps.get_model(\'main_app\', \'SomeItem\')\n\n# We can now test how our migration worked, new field is there:\nassert SomeItem.objects.filter(is_clean=True).count() == 0\nassert len(SomeItem._meta.get_fields()) == 3  # id + string_field + is_clean\n\n# Cleanup:\nmigrator.reset()\n```\n\nThat was an example of a forward migration.\n\n### Backward migration\n\nThe thing is that you can also test backward migrations.\nNothing really changes except migration names that you pass and your logic:\n\n```python\nmigrator = Migrator()\n\n# Currently our model has two field, but we need a rollback:\nold_state = migrator.before((\'main_app\', \'0002_someitem_is_clean\'))\nSomeItem = old_state.apps.get_model(\'main_app\', \'SomeItem\')\n\n# Create some data to illustrate your cases:\n# ...\n\n# Now this migration will drop `is_clean` field:\nnew_state = migrator.after((\'main_app\', \'0001_initial\'))\n\n# Assert the results:\n# ...\n\n# Cleanup:\nmigrator.reset()\n```\n\n### Testing migrations ordering\n\nSometimes we also want to be sure that our migrations are in the correct order.\nAnd all our `dependecies = [...]` are correct.\n\nTo achieve that we have [`plan.py`](https://github.com/wemake-services/django-test-migrations/blob/master/django_test_migrations/plan.py) module.\n\nThat\'s how it can be used:\n\n```python\nfrom django_test_migrations.plan import all_migrations, nodes_to_tuples\n\nmain_migrations = all_migrations(\'default\', [\'main_app\', \'other_app\'])\nassert nodes_to_tuples(main_migrations) == [\n    (\'main_app\', \'0001_initial\'),\n    (\'main_app\', \'0002_someitem_is_clean\'),\n    (\'other_app\', \'0001_initial\'),\n    (\'main_app\', \'0003_auto_20191119_2125\'),\n    (\'main_app\', \'0004_auto_20191119_2125\'),\n    (\'other_app\', \'0002_auto_20191120_2230\'),\n]\n```\n\nThis way you can be sure that migrations\nand apps that depend on each other will be executed in the correct order.\n\n\n## Test framework integrations\n\nWe support several test frameworks as first-class citizens.\nThat\'s a testing tool after all!\n\n### pytest\n\nWe ship `django-test-migrations` with a `pytest` plugin\nthat provides two convinient fixtures:\n\n- `migrator_factory` that gives you an opportunity\n  to create `Migrator` classes for any database\n- `migrator` instance for the `\'default\'` database\n\nThat\'s how it can be used:\n\n```python\nimport pytest\n\n@pytest.mark.django_db\ndef test_pytest_plugin_initial(migrator):\n    """Ensures that the initial migration works."""\n    old_state = migrator.before((\'main_app\', None))\n\n    with pytest.raises(LookupError):\n        # Models does not yet exist:\n        old_state.apps.get_model(\'main_app\', \'SomeItem\')\n\n    new_state = migrator.after((\'main_app\', \'0001_initial\'))\n    # After the initial migration is done, we can use the model state:\n    SomeItem = new_state.apps.get_model(\'main_app\', \'SomeItem\')\n    assert SomeItem.objects.filter(string_field=\'\').count() == 0\n```\n\n### unittest\n\nWe also ship an integration with the built-in `unittest` framework.\n\nHere\'s how it can be used:\n\n```python\nfrom django_test_migrations.contrib.unittest_case import MigratorTestCase\n\nclass TestDirectMigration(MigratorTestCase):\n    """This class is used to test direct migrations."""\n\n    migrate_from = (\'main_app\', \'0002_someitem_is_clean\')\n    migrate_to = (\'main_app\', \'0003_auto_20191119_2125\')\n\n    def prepare(self):\n        """Prepare some data before the migration."""\n        SomeItem = self.old_state.apps.get_model(\'main_app\', \'SomeItem\')\n        SomeItem.objects.create(string_field=\'a\')\n        SomeItem.objects.create(string_field=\'a b\')\n\n    def test_migration_main0003(self):\n        """Run the test itself."""\n        SomeItem = self.new_state.apps.get_model(\'main_app\', \'SomeItem\')\n\n        assert SomeItem.objects.count() == 2\n        assert SomeItem.objects.filter(is_clean=True).count() == 1\n```\n\n\n## Credits\n\nThis project is based on work of other awesome people:\n\n- [@asfaltboy](https://gist.github.com/asfaltboy/b3e6f9b5d95af8ba2cc46f2ba6eae5e2)\n- [@blueyed](https://gist.github.com/blueyed/4fb0a807104551f103e6#gistcomment-1546191)\n\n## License\n\nMIT.\n',
    'author': 'sobolevn',
    'author_email': 'mail@sobolevn.me',
    'url': 'https://github.com/wemake-services/django-test-migrations',
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
